import pathlib
import sys
import logging
import importlib

from .individual import simplify_this, AExpressionTree
from glyph.utils import Timeout

logger = logging.getLogger(__name__)


class Constraint:
    def __init__(self, spaces):
        self.spaces = spaces

    def _contains(self, element):
        if not self.spaces:
            return False
        return any(element in subspace for subspace in self.spaces)

    def __contains__(self, element):
        try:
            return self._contains(element)
        except Exception as e:
            logger.debug(f"Exception was raised during constraints check: {e}.")
            return False


class NonFiniteExpression(Constraint):
    def __init__(self, zero=True, infty=True, constant=False):
        """Use sympy to check for finite expressions.

        Args:
            zero: flag to check for zero expressions
            infty: flag to check for infinite expressions
            constant:  flag to check for constant expressions
        """
        self.zero = zero
        self.infty = infty
        self.constant = constant

    def _contains(self, element):
        expr = simplify_this(element)
        if isinstance(expr, str):
            logger.debug(f"Could not simplify {element}.")
            return False
        if self.constant:
            if expr.is_constant():
                return True
            elif all(t.name in element.pset.constants for t in element.terminals):
                return True

        if self.infty:
            if "zoo" in str(expr):
                return True

        if self.zero:
            if expr.is_zero:
                return True

        return False


class PreTest(Constraint):
    def __init__(self, fn, fun="chi"):
        """Apply pre-testing to check for constraint violation.

        The python script needs to provide a callable fun(ind).

        Args:
            fn: filename of the python script.
            fun: name of the function in fn.
        """
        fn = pathlib.Path(fn)
        sys.path.append(str(fn.parent))
        try:
            mod = importlib.import_module(fn.stem)
            self.f = getattr(mod, fun)
        except (AttributeError, ImportError):
            logger.error(f"Funktion {fun} not available in {fn}")
            self.f = lambda *args: False
        finally:
            sys.path.pop()

    def _contains(self, element):
        if self.f is None:
            logger.warning("Using invalid PretestNullSpace")
        return self.f(element)


class PreTestService(Constraint):  # todo com cannot be pickled !
    def __init__(self, assessment_runner):
        self.assessment_runner = assessment_runner

    @property
    def com(self):
        return self.assessment_runner.com

    @property
    def make_str(self):
        return self.assessment_runner.make_str

    def _contains(self, element):
        payload = self.make_str(element)
        self.com.send(dict(action="PRETEST", payload=payload))
        return self.com.recv()["ok"] == "True"


def reject_constrain_violation(constraint, n_trials=30, timeout=60):
    """Create constraints decorators based on rules.

    :param constraint:
    :param n_trials: Number of tries. Give up afterwards (return input).

    :return: list of constraint decorators
    """

    def reject(operator):
        def inner(*inds, **kw):
            with Timeout(timeout) as to_ctx:
                for i in range(n_trials):
                    out = operator(*inds, **kw)
                    if isinstance(out, AExpressionTree):  # can this be done w/o type checking?
                        t = out
                    elif isinstance(out, (list, tuple)):
                        t = out[0]
                    else:
                        raise RuntimeError
                    if t not in constraint:
                        break
                else:
                    logger.warning(f"Number of tries exhausted during constrained operation {operator} on individual {inds}.")  # noqa
                    if inds:
                        return inds
                    else:
                        raise UserWarning
            if to_ctx.state == to_ctx.TIMED_OUT:
                logger.warning(f"Timeout during constrained operation {operator} on individual {inds}.")
            return out
        return inner
    return reject


def constrain(funcs, constraint, n_trials=30, timeout=60):
    """Decorate a list of genetic operators with constraints.

    :param funcs: list of operators (mate, mutate, create)
    :param constraint: instance of Nullspace
    :return: constrained operators
    """

    if not constraint:
        return funcs
    return [reject_constrain_violation(constraint, n_trials=n_trials, timeout=timeout)(f)
            for f in funcs]



"""Some usefull classes/functions for the fitness assessment part in gp problems."""

import numpy as np
import scipy
import toolz
import functools
import warnings
import logging

import stopit
import deprecated

from glyph.gp.individual import _get_index
from glyph.utils import timeoutable


logger = logging.getLogger(__name__)


class SingleProcessFactory:
    map = map

    def __call__(self):
        return self


class AAssessmentRunner(object):
    def __init__(self, parallel_factory=SingleProcessFactory()):
        """Abstract runner for the (parallel) assessment of individuals in a population.

        Child classes have to at least override the measure() method, which might be
        executed in a different process or even on a remote machine depending on the
        parallelization scheme.
        Child classes may override the setup() method, which is executed once on
        object instantiation.
        Child classes may override the assign_fitness() method, which is executed in
        the main process. This can be usefull if you want to locally post-process
        the results of measure(), when collected from remote processes.

        :param parallel_factory: callable() -> obj, obj has to implement some
                                 kind of (parallel) map() method.
        """
        self.parallel_factory = parallel_factory
        self._parallel = None
        self.setup()

    def __call__(self, population):
        """Update the fitness of each individual in population that has an invalid fitness.

        :param population: a squence of individuals.
        """
        if self._parallel is None:
            self._parallel = self.parallel_factory()
        invalid = [p for p in population if not p.fitness.valid]
        fitnesses = self._parallel.map(self.measure, invalid)
        for ind, fit in zip(invalid, fitnesses):
            self.assign_fitness(ind, fit)
        return len(invalid)

    def __getstate__(self):
        """Modify pickling behavior for the class.

        All the attributes except 'parallel' can be pickled.
        """
        state = self.__dict__.copy()
        del state['_parallel']  # Remove the unpicklable parallelization scheme.
        return state

    def __setstate__(self, state):
        """Modify unpickling behavior for the class."""
        self.__dict__.update(state)  # Restore instance attributes.
        self._parallel = None  # Restoring the parallelization scheme happens when the functor is called.

    def setup(self):
        """Default implementation."""
        pass

    def assign_fitness(self, individual, fitness):
        """Assign a fitness value (as returned by self.measure()) to idividual.

        Default implementation.
        """
        individual.fitness.values = fitness

    def measure(self, individual):
        """Return a fitness value for individual."""
        raise NotImplementedError


def measure(*funcs, pre=toolz.identity, post=toolz.identity):
    """
    Combine several measurement functions into one.

    Optionaly do pre- and/or post-processing.

    :param funcs: a sequence of measure functions as returned by measure() (eg.
                   `callable(*a, **kw) -> tuple`), and/or single valued functions
                   (eg. `callable(*a, **kw)` -> numerical value).
    :param pre: some pre-processing function that is to be apllied on input
                *once* before passing the result to each function in funcs.
    :param post: some post-processing function that is to be apllied on the
                 tuple of measure values as returned by the combined funcs.
    :returns: callable(input) -> tuple of measure values, where input is usually
              a phenotype (eg. an expression tree).
    """
    def closure(*args, **kwargs) -> tuple:
        m = toolz.compose(post, _tt_flatten, toolz.juxt([tuple_wrap(f) for f in funcs]), pre)
        return m(*args, **kwargs)
    closure.size = sum(getattr(f, 'size', 1) for f in funcs)
    return closure


def default_constants(ind):
    """Return a one for each different constant in the primitive set.

    :param ind:
    :type ind: `glyph.gp.individual.AExpressionTree`
    :returns: A value for each constant in the primitive set.
    """
    if ind.pset.constants:
        consts_types = ind.pset.constants
        if len(consts_types) == 1 and "Symc" in consts_types:   # symc case
            values = np.ones(len(_get_index(ind, consts_types[0])))
        else:                           # sympy case
            values = np.ones(len(consts_types))
    else:
        values = []
    return values


def const_opt(measure, individual, lsq=False, default_constants=default_constants, f_kwargs=None, **kwargs):
    """Apply constant optimization

    :param measure: callable(individual, *f_args) -> scalar.
    :param individual: an individual tha is passed on to measure.
    :param bounds: bounds for the constant values (s. `scipy.optimize.minimize`).
    :param method: Type of solver. Should either be 'leastsq', or one of
             scipy.optimize.minimize's solvers.
    :returns: (popt, measure_opt), popt: the optimal values for the constants;
              measure_opt: the measure evaluated at popt.
    """

    opt = scipy.optimize.minimize if not lsq else scipy.optimize.least_squares
    f_kwargs = f_kwargs or {}

    @functools.wraps(measure)
    def closure(consts):
        return measure(individual, *consts, **f_kwargs)

    p0 = default_constants(individual)
    popt = p0
    measure_opt = None
    terminals = [t.name for t in individual.terminals]
    if any(constant in terminals for constant in individual.pset.constants):
        try:
            with warnings.catch_warnings():  # filter "unknown options for optimizer" warning
                warnings.filterwarnings("ignore", category=scipy.optimize.OptimizeWarning)
                res = opt(fun=closure, x0=p0, **kwargs)
            popt = res.x if res.x.shape else np.array([res.x])
            measure_opt = res.fun
            if not res.success:
                logger.debug(f"Evaluating {str(individual)}: {res.message}")
        except ValueError:
            return p0, closure(p0)
    if measure_opt is None:
        measure_opt = closure(popt)
    return popt, measure_opt


@deprecated.deprecated(reason="Use const_opt instead.", version="0.4")
def const_opt_scalar(*args, **kwargs):
    return const_opt(*args, **kwargs)


@deprecated.deprecated(reason="Use const_opt(*args, lsq=True, **kwargs) instead.", version="0.4")
def const_opt_leastsq(measure, individual, default_constants=default_constants, f_kwargs=None, **kwargs):
    return const_opt(measure, individual, lsq=True, default_constants=default_constants, f_kwargs=f_kwargs, **kwargs)


def replace_nan(x, rep=np.infty):
    """Replace occurences of np.nan in x.

    :param x: Any data structure
    :type x: list, tuple, float, np.ndarray
    :param rep: value to replace np.nan with

    :returns: x without nan's
    """
    @np.vectorize
    def repl(x):
        return rep if np.isnan(x) else x
    t = toolz.identity if isinstance(x, np.ndarray) else type(x)
    return t(repl(x))


def _tt_flatten(tt):
    """Flatten a tuple of tuples."""
    return tuple(sum(tt, ()))


def _argcount(func):
    """Return argument count of func."""
    return func.__code__.co_argcount


def annotate(func, annotations):
    """Add annoations to func."""
    if not hasattr(func, '__annoations__'):
        setattr(func, '__annotations__', annotations)
    else:
        func.__annotations__.update(annotations)
    return func


def returns(func, types):
    """Check func's annotation dictionary for return type tuple."""
    try:
        return func.__annotations__['return'] in types
    except (AttributeError, KeyError):
        return False


def tuple_wrap(func):
    """Wrap func's return value into a tuple if it is not one already."""
    types = tuple, list
    if returns(func, types=types):
        return func  # No need to wrap.
    else:
        @functools.wraps(func)
        def closure(*args, **kwargs):
            res = func(*args, **kwargs)
            if isinstance(res, types):
                return tuple(res)
            else:
                return res,
        annotate(closure, {'return': tuple})
        return closure


def max_fitness_on_timeout(max_fitness, timeout):
    """Decorate a function. Associate max_fitness with long running individuals.

    :param max_fitness: fitness of aborted individual calls.
    :param timeout: time until timeout
    :returns: fitness or max_fitness
    """
    def decorate(f):
        @functools.wraps(f)
        def inner(*args, **kwargs):
            return timeoutable(default=max_fitness)(f)(*args, timeout=timeout, **kwargs)
        return inner
    return decorate


def expressional_complexity(ind):
    """Sum of length of all subtrees of the individual."""
    return sum(len(ind[ind.searchSubtree(i)]) for i in range(len(ind)))


complexity_measures = {
    "ec_genotype": expressional_complexity,
    "ec_phenotype": lambda ind: expressional_complexity(ind.resolve_sc()),
    "num_nodes_genotype": len,
    "num_nodes_phenotype": lambda ind: len(ind.resolve_sc())
}



"""Provide Individual class for gp."""

import abc
import copy
import functools
import itertools
import logging
import re
import sys

import deap.base
import deap.gp
import numpy as np
import sympy

import glyph.utils

logger = logging.getLogger(__name__)


def sc_qout(x, y):
    """SC is the quotient of the number of nodes of its left and right child-trees x and y"""
    return x / y


def sc_mmqout(x, y, cmin=-1, cmax=1):
    """SC is the minimum-maximum quotient of the number of nodes of both
    child-trees x and y mapped into the constant interval [cmin, cmax]"""
    return cmin + min(x, y)/max(x, y) * (cmax - cmin)


class StructConst(deap.gp.Primitive):
    def __init__(self, func, arity=2):
        """
        :param func: evaluate left and right subtree and assign a constant.
        """
        self.func = func
        super().__init__("SC", [deap.gp.__type__]*arity, deap.gp.__type__)

    @staticmethod
    def get_len(expr, tokens=("(,")):
        regex = "|".join("\\{}".format(t) for t in tokens)
        return len(re.split(regex, expr))

    def as_terminal(self, *args):
        vals = list(map(len, args))
        return deap.gp.Terminal(self.func(*vals), False, deap.gp.__type__)


def add_sc(pset, func):
    """Adds a structural constant to a given primitive set.

    :param func: `callable(x, y) -> float` where x and y are the expressions of the left and right subtree
    :param pset: You may want to use `sympy_primitive_set` or `numpy_primitive_set` without symbolic constants.
    :type pset: `deap.gp.PrimitiveSet`
    """
    sc = StructConst(func)
    pset._add(sc)
    pset.prims_count += 1
    return pset


def _build_args_string(pset, consts):
    args = ','.join(arg for arg in pset.args)
    if consts:
        if args:
            args += ','
        args += ','.join("{}=1.0".format(arg) for arg in consts)
    return args


def sympy_primitive_set(categories=('algebraic', 'trigonometric', 'exponential'), arguments=['y_0'], constants=[]):
    """Create a primitive set with sympy primitves.

    :param arguments: variables to use in primitive set
    :param constants: symbolic constants to use in primitive set
    :param categories: an optional list of function categories for the primitive set. The following are available
                       'algebraic', 'neg', 'trigonometric', 'exponential', 'exponential', 'logarithm', 'sqrt'.

     :return: `deap.gp.PrimitiveSet`
    """
    pset = deap.gp.PrimitiveSet('main', arity=0)
    if 'algebraic' in categories:
        pset.addPrimitive(sympy.Add, arity=2)
        pset.addPrimitive(sympy.Mul, arity=2)
    if 'neg' in categories:
        pset.addPrimitive(functools.partial(sympy.Mul, -1.0), arity=1, name='Neg')
    if 'trigonometric' in categories:
        pset.addPrimitive(sympy.sin, arity=1)
        pset.addPrimitive(sympy.cos, arity=1)
    if 'exponential' in categories:
        pset.addPrimitive(sympy.exp, arity=1)
    if 'logarithm' in categories:
        pset.addPrimitive(sympy.ln, arity=1)
    if 'sqrt' in categories:
        pset.addPrimitive(sympy.sqrt, arity=1)

    # Use sympy symbols for argument representation.
    for symbol in itertools.chain(arguments, constants):
        pset.addTerminal(sympy.Symbol(symbol), name=symbol)

    pset.args = arguments
    pset.constants = constants
    return pset


def sympy_phenotype(individual):
    """Compile python function from individual.

    Uses sympy's lambdify function. Terminals from the primitive set will be
    used as parameters to the constructed lambda function; primitives (like
    sympy.exp) will be converted into numpy expressions (eg. numpy.exp).

    :param individual:
    :type individual: `glyph.gp.individual.AExpressionTree`
    :return: lambda function
    """
    pset = individual.pset
    args = _build_args_string(pset, pset.constants)
    expr = sympy.sympify(deap.gp.compile(str(individual), pset))
    func = sympy.utilities.lambdify(args, expr, modules='numpy')
    return func


def numpy_primitive_set(arity, categories=('algebraic', 'trigonometric', 'exponential', 'symc')):
    """Create a primitive set based on numpys vectorized functions.


    :param arity: Number of variables in the primitive set
    :param categories:
    :return: `deap.gp.PrimitiveSet`

    .. note :: All functions will be closed, that is non-defined values will be mapped to 1. 1/0 = 1!
    """
    pset = deap.gp.PrimitiveSet("main", arity)
    # Use primitive set built-in for argument representation.
    pset.renameArguments(**{'ARG{}'.format(i): 'x_{}'.format(i) for i in range(arity)})
    pset.args = pset.arguments
    if 'symc' in categories:
        symc = 1.0
        pset.addTerminal(symc, "Symc")
        pset.constants = ["Symc"]
    else:
        pset.constants = []

    def close_function(func, value):
        @functools.wraps(func)
        @glyph.utils.numeric.silent_numpy
        def closure(*args):
            res = func(*args)
            if isinstance(res, np.ndarray):
                res[np.isinf(res)] = value
                res[np.isnan(res)] = value
            elif np.isinf(res) or np.isnan(res):
                res = value
            return res
        return closure

    if 'algebraic' in categories:
        pset.addPrimitive(np.add, 2, name="Add")
        pset.addPrimitive(np.subtract, 2, name="Sub")
        pset.addPrimitive(np.multiply, 2, name="Mul")
        _div = close_function(np.divide, 1)
        pset.addPrimitive(_div, 2, name="Div")
    if 'trigonometric' in categories:
        pset.addPrimitive(np.sin, arity=1)
        pset.addPrimitive(np.cos, arity=1)
    if 'exponential' in categories:
        pset.addPrimitive(np.exp, arity=1)
        _log = close_function(np.log, 1)
        pset.addPrimitive(_log, arity=1)
    if 'sqrt' in categories:
        _sqrt = close_function(np.sqrt, 1)
        pset.addPrimitive(_sqrt, arity=1)
    if "constants" in categories:
        pset.addTerminal(np.pi, name="pi")
        pset.addTerminal(np.e, name="e")
    return pset


def _get_index(ind, c):
    return [i for i, node in enumerate(ind) if node.name == c]


def numpy_phenotype(individual):
    """ Lambdify the individual

    :param individual:
    :type individual: `glyph.gp.individual.AExpressionTree`
    :return: lambda function

    :Note:
    In constrast to sympy_phenotype the callable will have a variable number of keyword arguments depending on the number
    of symbolic constants in the individual.

    :Example:
    >>> pset = numpy_primitive_set(1)
    >>> MyIndividual = Individual(pset=pset)
    >>> ind = MyIndividual.from_string("Add(x_0, Symc)")
    >>> f = numpy_phenotype(ind)
    >>> f(1, 1)
    2
    """
    pset = individual.pset
    if pset.constants:
        c = pset.constants[0]
        index = _get_index(individual, c)
    else:
        index = []
    consts = ["c_{}".format(i) for i in range(len(index))]
    args = _build_args_string(pset, consts)
    expr = str(individual)
    for c_ in consts:
        expr = expr.replace(c, c_, 1)
    func = sympy.utilities.lambdify(args, expr, modules=pset.context)
    return func


class Individual(type):
    """Metaclass to set the primitive set in ExpressionTree types."""

    def __new__(mcs, pset, name="MyIndividual", **kwargs):
        """Construct a new expression tree type.

        Args:
            pset: :class:`deap.gp.PrimitiveSet`
            name: name of the expression tree class
            kwargs: additional attributes

        Returns: expression tree class
        """
        cls = super().__new__(mcs, name, (AExpressionTree,), dict(pset=pset, **kwargs))
        setattr(sys.modules[__name__], name, cls)
        return cls

    def __init__(cls, pset, name="MyIndividual", **kwargs):  # noqa
        super().__init__(name, (AExpressionTree,), dict(pset=pset, **kwargs))


class AExpressionTree(deap.gp.PrimitiveTree):
    hasher = str

    def __init__(self, content):
        """Abstract base class for the genotype.

        Derived classes need to specify a primitive set from which the expression
        tree can be build, as well as a phenotype method.
        """
        super(AExpressionTree, self).__init__(content)
        self.fitness = Measure()

    def __repr__(self):
        return self.to_polish(replace_struct=False)

    def __str__(self):
        return self.to_polish()

    def resolve_sc(self):
        """Evaluate StructConst in individual top to bottom."""
        nodes = type(self)(self[:])

        # structure based constants need to be evaluated top to bottom first
        is_sc = lambda n: isinstance(n, StructConst)
        while any(n for n in nodes if is_sc(n)):
            # get the first structure based constant
            i, node = next(filter((lambda x: is_sc(x[1])), enumerate(nodes)))
            slice_ = nodes.searchSubtree(i)
            # find its subtree
            subtree = type(self)(nodes[slice_])
            # replace the subtree by a Terminal representing its numeric value
            # child_trees helper function yields a list of tree which are used as arguments by the first primitive
            args = list(child_trees(subtree))
            nodes[slice_] = [node.as_terminal(*args)]
        return nodes

    def to_polish(self, for_sympy=False, replace_struct=True):
        """Symbolic representation of the expression tree."""
        nodes = self.resolve_sc() if replace_struct else self
        repr = ""
        stack = []
        for node in nodes:
            stack.append((node, []))
            while len(stack[-1][1]) == stack[-1][0].arity:
                prim, args = stack.pop()
                repr = prim.format(*args) if not for_sympy else convert_inverse_prim(prim, args)
                if len(stack) == 0:
                    break
                stack[-1][1].append(repr)
        return repr

    def __hash__(self):
        return hash(self.hasher(self))

    def __eq__(self, other):
        return hash(self) == hash(other)

    @property
    def const_opt(self):
        return getattr(self, "popt", ())

    @property
    def terminals(self):
        """Return terminals that occur in the expression tree."""
        return [primitive for primitive in self if primitive.arity == 0]

    @property
    @abc.abstractmethod
    def pset(self):
        pass

    @classmethod
    def from_string(cls, string):
        return super(AExpressionTree, cls).from_string(string, cls.pset)

    @classmethod
    def create(cls, gen_method=deap.gp.genHalfAndHalf, min=1, max=4):
        return cls(gen_method(cls.pset, min_=min, max_=max))

    @classmethod
    def create_population(cls, size, gen_method=deap.gp.genHalfAndHalf, min=1, max=4):
        """Create a list of individuals of class Individual."""
        if size < 0:
            raise RuntimeError('Cannot create population of size {}'.format(size))
        return [cls.create(gen_method=gen_method, min=min, max=max) for _ in range(size)]


def nd_phenotype(nd_tree, backend=sympy_phenotype):
    """
    :param nd_tree:
    :type  nd_tree: ANDimTree
    :param backend: sympy_phenotype or numpy_phenotype
    :return: lambda function
    """
    funcs = [backend(t) for t in nd_tree]
    return lambda *x: [f(*x) for f in funcs]


class NDIndividual(type):
    def __new__(mcs, base, name="MyNDIndividual", **kwargs):
        cls = super().__new__(mcs, name, (ANDimTree,), dict(base=base, **kwargs))
        setattr(sys.modules[__name__], name, cls)
        return cls

    def __init__(cls, base, name="MyNDIndividual", **kwargs):  # noqa
        """Construct a new n-dimensional expression tree type.

        Args:
            base (Individual): one dimensional base class
            name: name of the n-dimensional expression tree class
            **kwargs: addtional attributes

        Returns: n-dimensional expression tree class
        """
        super().__init__(name, (ANDimTree,), dict(base=base, **kwargs))


class ANDimTree(list):
    """ A naive tree class representing a vector-valued expression.

    Each dimension is encoded as a expression tree.
    """
    def __init__(self, trees):
        super().__init__(trees)
        self.dim = len(trees)
        self.fitness = Measure()

    @property
    @abc.abstractmethod
    def base(self):
        pass

    def __repr__(self):
        return str([str(tree) for tree in self])

    @classmethod
    def create_individual(cls, ndim):
        return cls(cls.base.create_population(ndim))

    @classmethod
    def create_population(cls, size, ndim):
        return [cls.create_individual(ndim) for _ in range(size)]

    @property
    def height(self):
        return len(self)

    @property
    def pset(self):
        return self.base.pset

    @property
    def terminals(self):
        """Return terminals that occur in the expression tree."""
        return [primitive for primitive in itertools.chain.from_iterable(self) if primitive.arity == 0]

    @classmethod
    def from_string(cls, strs):
        return cls([cls.base.from_string(s) for s in strs])


class Measure(deap.base.Fitness):
    """
    This is basically a wrapper around deap.base.Fitness.

    It provides the following enhancements over the base class:
    - more adequate naming
    - copy constructable
    - no weight attribute
    """

    weights = itertools.repeat(-1)

    def __init__(self, values=()):
        if len(values) > 0:
            self.values = values

    def __iter__(self):
        return iter(self.values)

    def __len__(self):
        return len(self.values)

    def get_values(self):
        return tuple(-1.0 * v for v in self.wvalues)

    def set_values(self, values):
        self.wvalues = tuple(-1.0 * v for v in values)

    def del_values(self):
        self.wvalues = ()

    values = property(get_values, set_values, del_values)


def convert_inverse_prim(prim, args):
    """
    Convert inverse prims according to:
    [Dd]iv(a,b) -> Mul[a, 1/b]
    [Ss]ub(a,b) -> Add[a, -b]

    We achieve this by overwriting the corresponding format method of the sub and div prim.
    """
    prim = copy.copy(prim)
    converter = {
        'sub': lambda *args_: "Add({}, Mul(-1,{}))".format(*args_),
        'div': lambda *args_: "Mul({}, Pow({}, -1))".format(*args_),
        'add': lambda *args_: "Add({}, {})".format(*args_),
        'mul': lambda *args_: "Mul({}, {})".format(*args_),
    }
    prim_formatter = converter.get(prim.name.lower(), prim.format)
    return prim_formatter(*args)


@glyph.utils.Memoize
def simplify_this(expr, timeout=5):
    """
    :param expr:
    :type expr: str
    :return: Sympy representation of simplified expression

    :warning: does not respect closures
    """
    with glyph.utils.random_state(simplify_this):  # to avoid strange side effects of sympy testcases and random
        with glyph.utils.Timeout(timeout):
            try:
                return sympy.simplify(expr.to_polish(for_sympy=True))
            except Exception as e:
                logger.debug(f"Exception during simplification of {expr}: {e}.")
                return expr
        return expr


def child_trees(ind):
    """Yield all child tree which are used as arguments for the head node of ind."""
    start = 1
    while start < len(ind):
        slice_ = ind.searchSubtree(start)
        yield type(ind)(ind[slice_])
        start += slice_.stop - 1


@glyph.utils.Memoize
def simplify_constant(ind):
    """Trims subtrees of symbolic constants.

    Recursively applying these rules:

        S f(*[symc]*f.arity) = symc
        S f(x,... symc) = f(x,... symc)
        S symc = symc

    For deeper trees, try to trim down lower levels first.
    An individual cannot be trimmed down further if its a fixpoint of S.
    """
    symc = ind.pset.mapping[ind.pset.constants[0]]
    if symc in ind:
        root = ind.root

        # the tree is just a function of constants
        if len(ind) > 1 and  all(i == symc for i in ind[1:]):
            return type(ind)([symc])

        # root is just a constant or tree is a function of a variable and cannot be trimmed down.
        elif len(ind) == 1 or len(ind) == root.arity + 1:
            return ind

        # try to simplify all children of root
        else:
            acc = [simplify_constant(child) for child in child_trees(ind)]
            new_ind = type(ind)([root] + sum(acc, []))

            # cannot be simplified
            if ind == new_ind:
                return ind

            # can it be simplified further?
            else:
                return simplify_constant(new_ind)

    else:
        return ind


def _constant_normal_form(expr, variables=()):
    """experimental
    """
    if expr.is_constant(*variables):
        return sympy.Symbol("c")

    args = expr.args
    if not args:
        return expr
    elif isinstance(expr, sympy.Pow):
        return type(expr)(*[_constant_normal_form(a, variables=variables) for a in args[:-1]] + [args[-1]])
    else:
        res = type(expr)(*[_constant_normal_form(a, variables=variables) for a in args])
        if res == expr:
            return expr
        else:
            return _constant_normal_form(res, variables=variables)


def pretty_print(expr, constants, consts_values, count=0):
    """Replace symbolic constants in the str representation of an individual
    by their numeric values.

    This checks for
        - c followed by ")" or ","
        - c followed by infix operators
        - c
    """
    for k, v in zip(constants, consts_values):
        c = str(k)
        p1 = r"{c}(?=[,)])".format(c=c)
        p2 = r"^{c}$".format(c=c)
        p3 = r"(?<=[*+-/]){c}|(?<=[*+-/] ){c}|{c}(?=\s?[*+-/])".format(c=c)
        pattern = r"|".join((p1, p2, p3))
        expr = re.sub(pattern, str(v), expr, count=count)
    return expr



import os
import random
import itertools
from contextlib import contextmanager

import stopit

from . import argparse
from . import logging
from . import numeric
from . import break_condition


class Memoize:
    def __init__(self, fn):
        """Memoize(fn) - an instance which acts like fn but memoizes its arguments

        Will only work on functions with non-mutable arguments
        http://code.activestate.com/recipes/52201/
        """
        self.fn = fn
        self.memo = {}

    def __call__(self, *args, **kwargs):
        key = str(args) + str(kwargs)
        if key not in self.memo:
            self.memo[key] = self.fn(*args, **kwargs)
        return self.memo[key]


@contextmanager
def random_state(obj, rng=random):
    """Do work inside this contextmanager with a random state defined by obj.

    Looks for _prev_state to seed the rng.
    On exit, it will write the current state of the rng as _tmp_state
    to the obj.

    :params obj: Any object.
    :params rng: Instance of a random number generator.
    """
    obj._tmp_state = rng.getstate()
    rng.setstate(getattr(obj, '_prev_state', rng.getstate()))
    yield
    obj._prev_state = rng.getstate()
    rng.setstate(getattr(obj, '_tmp_state', rng.getstate()))


def partition(pred, iterable):
    """Use a predicate to partition entries into false entries and true entries.

    >>> is_odd = lambda x: x % 2
    >>> odd, even = partition(is_odd, range(10))
    >>> list(odd)
    [0, 2, 4, 6, 8]
    """
    t1, t2 = itertools.tee(iterable)
    return itertools.filterfalse(pred, t1), filter(pred, t2)


def key_set(itr, key=hash):
    keys = map(key, itr)
    s = {k: v for k, v in zip(keys, itr)}
    return list(s.values())


_is_posix = "posix" in os.name
Timeout = stopit.SignalTimeout if _is_posix else stopit.ThreadingTimeout
timeoutable = stopit.signal_timeoutable if _is_posix else stopit.threading_timeoutable



import concurrent.futures
import copy
import enum
import itertools
import json
import logging
import os
import random
from functools import partial
from queue import Queue
from threading import Thread
from time import sleep

logging.captureWarnings(True)

import yaml

import deap.gp
import deap.tools
import sympy
import zmq
from deprecated import deprecated
from cache import DBCache
from scipy.optimize._minimize import _minimize_neldermead as nelder_mead


import glyph.application
import glyph.gp.individual
import glyph.utils
from glyph._version import get_versions
from glyph.assessment import const_opt, complexity_measures
from glyph.cli._parser import *  # noqa
from glyph.gp.constraints import constrain
from glyph.gp.individual import _constant_normal_form, add_sc, pretty_print, sc_mmqout, simplify_this
from glyph.observer import ProgressObserver
from glyph.utils import partition, key_set
from glyph.utils.argparse import *  # noqa
from glyph.utils.break_condition import break_condition
from glyph.utils.logging import print_params, load_config

logger = logging.getLogger(__name__)


def get_version_info():
    version = get_versions()["version"]
    return f"Version {version}"


class ExperimentProtocol(enum.EnumMeta):
    """Communication Protocol with remote experiments."""

    CONFIG = "CONFIG"
    EXPERIMENT = "EXPERIMENT"
    METADATA = "METADATA"
    SHUTDOWN = "SHUTDOWN"


class Communicator:
    def __init__(self, ip, port):
        """Holds the socket for 0mq communication.

        Args:
            ip: ip of the client
            port: port of the client
        """
        self._socket = zmq.Context.instance().socket(zmq.REQ)
        self.ip = ip
        self.port = port

    def connect(self):
        address = f"tcp://{self.ip}:{self.port}"
        logger.debug(f"Connecting to experiment on {address}")
        self._socket.connect(address)

    def send(self, msg, serializer=json):
        logger.log(logging.NOTSET, msg)
        self._socket.send(serializer.dumps(msg).encode("ascii"))

    def recv(self, serializer=json):
        msg = serializer.loads(self._socket.recv().decode("ascii"))
        logger.log(logging.NOTSET, msg)
        return msg


class RemoteApp(glyph.application.Application):
    def run(self, break_condition=None):
        """For details see application.Application.
        Will checkpoint and close zmq connection on keyboard interruption.
        """
        try:
            super().run(break_condition=break_condition)
        except KeyboardInterrupt:
            logger.info("Received KeyboardInterrupt. Trying checkpointing.")
            self.checkpoint()
        finally:
            self.assessment_runner.com.send(dict(action=ExperimentProtocol.SHUTDOWN))
            zmq.Context.instance().destroy()

    @classmethod
    def from_checkpoint(cls, file_name, com):
        """Create application from checkpoint file."""
        cp = glyph.application.load(file_name)
        args = cp["args"]
        gp_runner = cp["runner"]
        gp_runner.assessment_runner = RemoteAssessmentRunner(
            com,
            complexity_measure=args.complexity_measure,
            method=args.const_opt_method,
            options=args.options,
            caching=args.caching,
            simplify=args.simplify,
            persistent_caching=args.persistent_caching,
            chunk_size=args.chunk_size,
            multi_objective=args.multi_objective,
            send_symbolic=args.send_symbolic,
            reevaluate=args.re_evaluate,
        )
        app = cls(args, gp_runner, file_name, cp["callbacks"])
        app.pareto_fronts = cp["pareto_fronts"]
        app._initialized = True
        pset = build_pset_gp(args.primitives, args.structural_constants, args.sc_min, args.sc_max)
        Individual.pset = pset
        random.setstate(cp["random_state"])
        return app

    def checkpoint(self):
        """Checkpoint current state of evolution."""
        runner = copy.deepcopy(self.gp_runner)
        del runner.assessment_runner
        glyph.application.safe(
            self.checkpoint_file,
            args=self.args,
            runner=runner,
            random_state=random.getstate(),
            pareto_fronts=self.pareto_fronts,
            callbacks=self.callbacks,
        )


def handle_const_opt_config(args):
    smart_options = {
        "use": args.smart,
        "kw": {
            "threshold": args.smart_threshold,
            "step_size": args.smart_step_size,
            "min_stat": args.smart_min_stat,
        },
    }
    options = {"maxfev": args.max_fev_const_opt, "smart_options": smart_options}
    if args.const_opt_method == "hill_climb":
        options["directions"] = args.directions
        options["precision"] = args.precision
        options["target"] = args.target
    else:
        options["xatol"] = 10.0 ** (-args.precision)
        options["fatol"] = args.target
    args.options = options
    return args


def const_opt_options_transform(options):
    leastsq_options = dict()
    leastsq_options["xtol"] = options["xatol"]
    leastsq_options["ftol"] = options["fatol"]
    leastsq_options["max_nfev"] = options["maxfev"]
    return leastsq_options


def update_namespace(ns, up):
    """Update the argparse.Namespace ns with a dictionairy up."""
    return argparse.Namespace(**{**vars(ns), **up})


def handle_gpconfig(config, com):
    """Will try to load config from file or from remote and update the cli/default config accordingly."""
    if config.cfile:
        with open(config.cfile, "r") as cf:
            gpconfig = yaml.load(cf)
    elif config.remote:
        com.send(dict(action=ExperimentProtocol.CONFIG))
        gpconfig = com.recv()
    else:
        gpconfig = {}
    return update_namespace(config, gpconfig)


def build_pset_gp(primitives, structural_constants=False, cmin=-1, cmax=1):
    """Build a primitive set used in remote evaluation.

    Locally, all primitives correspond to the id() function.
    """

    if any(x in primitives for x in ["add", "mul"]):
        logger.warning("Some primitive names may be incompatible with sympy related features.")

    pset = deap.gp.PrimitiveSet("main", arity=0)
    pset.constants = set()
    for fname, arity in primitives.items():
        if arity > 0:
            func = lambda *args: args
            pset.addPrimitive(func, arity, name=fname)
        elif arity == 0:
            pset.addTerminal(fname, name=fname)
            pset.arguments.append(fname)
        else:
            pset.addTerminal(fname, name=fname)
            pset.constants.add(fname)
    if len(pset.terminals) == 0:
        raise RuntimeError("Pset needs at least one terminal node. You may have forgotten to specify it.")

    if structural_constants:
        f = partial(sc_mmqout, cmin=cmin, cmax=cmax)
        pset = add_sc(pset, f)
    return pset


class EvalQueue(Queue):
    def __init__(self, com, result_queue, expect):
        self.com = com
        self.result_queue = result_queue
        self.expect = expect

        super().__init__()

    def run(self, chunk_size=100):
        payloads = []
        keys = []

        def process(keys, payload_meta):
            payload, meta = zip(*payload_meta)
            logger.debug(f"Send {len(payload)} payloads.")
            if any(meta):
                self.com.send(dict(action=ExperimentProtocol.EXPERIMENT, payload=payload, meta=meta))
            else:
                self.com.send(dict(action=ExperimentProtocol.EXPERIMENT, payload=payload))
            fitnesses = self.com.recv()["fitness"]
            logger.debug(f"Received {len(fitnesses)} fitness values.")
            for key, fit in zip(keys, fitnesses):
                logger.log(logging.NOTSET, "Writing result for key: {}".format(key))
                self.result_queue[key] = fit

        while self.expect > 0:
            key_payload_meta = self.get()

            if key_payload_meta is None:
                self.expect -= 1
                if self.expect == 0:
                    break
            else:
                key, payload_meta = key_payload_meta
                if key not in self.result_queue:
                    logger.log(logging.NOTSET, "Queueing key: {}".format(key))
                    payloads.append(payload_meta)
                    keys.append(key)
            if len(payloads) == min(self.expect, chunk_size):
                process(keys, payloads)
                payloads = []
                keys = []
        if payloads:
            process(keys, payloads)


def _no_const_opt(func, ind):
    return None, func(ind)


class RemoteAssessmentRunner:
    def __init__(
        self,
        com,
        complexity_measure=None,
        multi_objective=False,
        method="Nelder-Mead",
        options={"smart_options": {"use": False}},
        caching=True,
        persistent_caching=None,
        simplify=False,
        chunk_size=30,
        send_symbolic=False,
        reevaluate=False,
    ):
        """Contains assessment logic. Uses zmq connection to request evaluation."""
        self.com = com
        self.complexity_measure = complexity_measures.get(complexity_measure, None)
        self.multi_objective = multi_objective
        self.caching = caching
        self.cache = {} if persistent_caching is None else DBCache("glyph-remote", persistent_caching)
        self.make_str = (lambda i: str(simplify_this(i))) if simplify else str
        self.result_queue = {}
        self.send_symbolic = send_symbolic
        self.reevaluate = reevaluate
        self.evaluations = 0
        self.chunk_size = chunk_size
        if chunk_size > 30:
            logger.warning("Chunk size may cause performance issues.")

        if not self.send_symbolic:
            self.options = options
            self.method = {"hill_climb": glyph.utils.numeric.hill_climb}.get(method, nelder_mead)

            self.smart_options = options.get("smart_options")
            if self.smart_options["use"]:
                self.method = glyph.utils.numeric.SlowConversionTerminator(
                    glyph.utils.numeric.hill_climb, **self.smart_options["kw"]
                )

            if self.multi_objective:
                self.const_optimizer = partial(
                    const_opt, lsq=True, **const_opt_options_transform(self.options)
                )
            else:
                self.const_optimizer = partial(const_opt, method=self.method, options=self.options)

        else:
            self.const_optimizer = _no_const_opt

    def predicate(self, ind):
        """Does this individual need to be evaluated?"""
        return self.caching and self._hash(ind) in self.cache

    def _hash(self, ind):
        return json.dumps([self.make_str(t) for t in ind])

    def evaluate_single(self, individual, *consts, meta=None):
        """Evaluate a single individual."""
        payload = [self.make_str(t) for t in individual]
        if not self.send_symbolic:
            payload = [pretty_print(s, individual.pset.constants, consts) for s in payload]
            key = sum(map(hash, payload))  # constants may have been simplified, not in payload anymore.
        else:
            variables = [sympy.Symbol(s) for s in Individual.pset.arguments]
            normal_form = [_constant_normal_form(sympy.sympify(p), variables=variables) for p in payload]
            key = sum(map(hash, normal_form))

        self.queue.put((key, (payload, meta)))
        self.evaluations += 1

        result = None
        while result is None:
            sleep(0.1)
            result = self.result_queue.get(key)
        return result

    def measure(self, individual, meta=None):
        """Construct fitness for given individual."""
        popt, error = self.const_optimizer(self.evaluate_single, individual, f_kwargs=dict(meta=meta))
        if not self.multi_objective:
            error = (error,)

        self.queue.put(None)
        individual.popt = popt
        if self.complexity_measure:
            fitness = *error, sum(map(self.complexity_measure, individual))
        else:
            fitness = error
        return fitness

    def update_fitness(self, population, meta=None):
        self.evaluations = 0
        meta = meta or {}

        if self.reevaluate:
            for p in population:
                del p.fitness.values

        invalid = [p for p in population if not p.fitness.valid]
        calculate, cached = map(list, partition(self.predicate, invalid))
        cached_fitness = [self.cache[self._hash(ind)] for ind in cached]
        calculate_duplicate_free = key_set(calculate, key=self._hash)
        # if we have duplicates in the calculate list, dont calculate these more than once.
        dup_free_cache = {}
        n = len(calculate_duplicate_free)

        if n > 0:  # main work is done here
            n_workers = min(n, self.chunk_size)

            # start queue and the broker
            self.queue = EvalQueue(self.com, self.result_queue, n)
            thread = Thread(target=self.queue.run, args=(n_workers,))
            thread.start()

            with concurrent.futures.ThreadPoolExecutor(max_workers=n_workers) as exe:
                for k, future in zip(
                    calculate_duplicate_free,
                    exe.map(partial(self.measure, meta=meta), calculate_duplicate_free),
                ):
                    dup_free_cache[self._hash(k)] = future
            thread.join()
            del self.queue

            calculate_fitness = [dup_free_cache[self._hash(k)] for k in calculate]
        else:
            calculate_fitness = []

        if self.caching:
            # save to cache
            for key, fit in zip(map(self._hash, calculate), calculate_fitness):
                self.cache[key] = fit

        # assign fitness to individuals
        for ind, fit in zip(cached + calculate, cached_fitness + calculate_fitness):
            ind.fitness.values = fit

        if self.reevaluate or not self.caching:
            self.result_queue = {}

        return self.evaluations

    def __call__(self, population, meta=None):
        meta = meta or {}
        return self.update_fitness(population, meta=meta)

    @property
    @deprecated(reason="Use RemoteAssessmentRunner.com.send instead.", version="0.3.7")
    def send(self):
        """Backwards compatibility"""
        return self.com.send

    @property
    @deprecated(reason="Use RemoteAssessmentRunner.com.recv instead.", version="0.3.7")
    def recv(self):
        """Backwards compatibility"""
        return self.com.recv


class Individual(glyph.gp.individual.AExpressionTree):
    pass


class NDTree(glyph.gp.individual.ANDimTree):
    base = Individual

    def __hash__(self):
        return hash(hash(x) for x in self)


def make_callback(factories, args):
    return tuple(factory(args) for factory in factories)


def make_remote_app(callbacks=(), callback_factories=(), parser=None):
    parser = parser or get_parser()
    args, _ = parser.parse_known_args()
    if isinstance(parser, Parser):
        if hasattr(args, "gui") and args.gui:
            if GUI_AVAILABLE:
                parser = get_parser(get_gooey())
            else:
                raise ValueError(GUI_UNAVAILABLE_MSG)
    args = parser.parse_args()
    com = Communicator(args.ip, args.port)
    com.connect()
    workdir = os.path.dirname(os.path.abspath(args.checkpoint_file))
    if not os.path.exists(workdir):
        raise RuntimeError('Path does not exist: "{}"'.format(workdir))

    glyph.utils.logging.load_config(
        config_file=args.logging_config,
        level=getattr(logging, args.verbosity),
        placeholders=dict(workdir=workdir)
    )
    if args.resume_file is not None:
        logger.debug("Loading checkpoint {}".format(args.resume_file))
        app = RemoteApp.from_checkpoint(args.resume_file, com)
    else:
        args = handle_const_opt_config(handle_gpconfig(args, com))
        try:
            pset = build_pset_gp(args.primitives, args.structural_constants, args.sc_min, args.sc_max)
        except AttributeError:
            raise AttributeError("You need to specify the pset")

        assessment_runner = RemoteAssessmentRunner(
            com,
            method=args.const_opt_method,
            options=args.options,
            complexity_measure=args.complexity_measure,
            caching=args.caching,
            persistent_caching=args.persistent_caching,
            simplify=args.simplify,
            chunk_size=args.chunk_size,
            multi_objective=args.multi_objective,
            send_symbolic=args.send_symbolic,
            reevaluate=args.re_evaluate,
        )

        Individual.pset = pset
        mate = glyph.application.MateFactory.create(args, Individual)
        mutate = glyph.application.MutateFactory.create(args, Individual)
        select = glyph.application.SelectFactory.create(args)
        create_method = glyph.application.CreateFactory.create(args, Individual)

        mate, mutate, Individual.create = constrain(
            [mate, mutate, Individual.create],
            glyph.application.ConstraintsFactory.create(args),
            n_trials=args.constraints_n_retries,
            timeout=args.constraints_timeout,
        )

        ndmate = partial(glyph.gp.breeding.nd_crossover, cx1d=mate)
        ndmutate = partial(glyph.gp.breeding.nd_mutation, mut1d=mutate)
        ndcreate = lambda size: [NDTree(create_method(args.ndim)) for _ in range(size)]
        NDTree.create_population = ndcreate
        algorithm_factory = partial(
            glyph.application.AlgorithmFactory.create, args, ndmate, ndmutate, select, ndcreate
        )

        gp_runner = glyph.application.GPRunner(NDTree, algorithm_factory, assessment_runner)

        callbacks = glyph.application.DEFAULT_CALLBACKS + callbacks + make_callback(callback_factories, args)

        if args.send_meta_data:
            callbacks += (send_meta_data,)

        if args.animate:
            callbacks += (ProgressObserver(),)

        app = RemoteApp(args, gp_runner, args.checkpoint_file, callbacks=callbacks)

    bc = break_condition(ttl=args.ttl, target=args.target, max_iter=args.max_iter_total, error_index=0)
    return app, bc, args


def log_info(args):
    logger.info(f"Glyph-remote. {get_version_info()}")
    logger.debug("Parameters:")
    print_params(logger.debug, vars(args))


def send_meta_data(app):
    com = app.gp_runner.assessment_runner.com

    metadata = dict(generation=app.gp_runner.step_count)
    com.send(dict(action=ExperimentProtocol.METADATA, payload=metadata))
    logger.debug(com.recv())


def main():
    app, bc, args = make_remote_app()
    log_info(args)
    app.run(break_condition=bc)


if __name__ == "__main__":
    main()



import copy
import sys
import random
import itertools
import deap.tools
import deap.algorithms


def _all_valid(population):
    """Checks if all individuals in a population have been evaluated"""
    return all(ind.fitness.valid for ind in population)


def varOr(population, toolbox, lambda_, cxpb, mutpb):
    """Just a fixed version of deap.algorithm.varOr
    Reproduction needs cloning, so that fitness is not shared.
    Else the current implementation of AgeFitness will break.
    """
    assert (cxpb + mutpb) <= 1.0, "The sum of the crossover and mutation probabilities must be smaller or equal to 1.0."

    offspring = []
    for _ in range(lambda_):
        op_choice = random.random()
        if op_choice < cxpb:            # Apply crossover
            ind1, ind2 = list(map(toolbox.clone, random.sample(population, 2)))
            ind1, ind2 = toolbox.mate(ind1, ind2)
            del ind1.fitness.values
            offspring.append(ind1)
        elif op_choice < cxpb + mutpb:  # Apply mutation
            ind = toolbox.clone(random.choice(population))
            ind, = toolbox.mutate(ind)
            del ind.fitness.values
            offspring.append(ind)
        else:                           # Apply reproduction
            offspring.append(toolbox.clone(random.choice(population)))

    return offspring


class MOGP(object):
    def __init__(self, mate_func, mutate_func, select):
        self.mate = mate_func
        self.mutate = mutate_func
        self.select = select
        self.clone = copy.deepcopy
        self.crossover_prob = 0.7
        self.mutation_prob = 0.2
        self.tournament_size = 2
        self._initialized = False

    def evolve(self, population):
        if not _all_valid(population):
            raise RuntimeError('Cannot evolve on invalid fitness values in population.')
        if not self._initialized:
            self._init(population)
        parents = self.select(population, self.parents_size)
        offspring = self._breed(parents)
        return parents[:] + offspring

    def _init(self, population):
        self.parents_size = len(population)
        self.offspring_size = len(population)
        self._initialized = True

    def _breed(self, parents):
        mating_pool = deap.tools.selTournament(parents, self.offspring_size, self.tournament_size)
        return varOr(mating_pool, self, self.offspring_size, self.crossover_prob, self.mutation_prob)


class NSGA2(MOGP):
    """Implementation of the NSGA-II algorithm as described in Essentials of Metaheuristics"""
    def __init__(self, mate_func, mutate_func):
        super().__init__(mate_func, mutate_func, deap.tools.selNSGA2)


class SPEA2(MOGP):
    """Implementation of the SPEA2 algorithm as described in Essentials of Metaheuristics"""
    def __init__(self, mate_func, mutate_func):
        super().__init__(mate_func, mutate_func, deap.tools.selSPEA2)


class DeapEaSimple(object):
    """Basically a copy of `deap.algorithms.eaSimple` algorithm."""

    def __init__(self, mate_func, mutate_func):
        self.mate = mate_func
        self.mutate = mutate_func
        self.clone = copy.deepcopy
        self.crossover_prob = 0.5
        self.mutation_prob = 0.2
        self.tournament_size = 2
        self._initialized = False

    def _init(self, population):
        self.parents_size = len(population)
        self.offspring_size = len(population)
        self._initialized = True

    def evolve(self, population):
        if not _all_valid(population):
            raise RuntimeError('Cannot evolve on invalid fitness values in population.')
        if not self._initialized:
            self._init(population)
        parents = deap.tools.selTournament(population, self.offspring_size, self.tournament_size)
        offspring = deap.algorithms.varAnd(parents, self, self.crossover_prob, self.mutation_prob)  # Breeding.
        return population[:] + offspring


class AgeFitness(MOGP):
    """AgeFitness algorithm as described in Schmidt & Lipson.
    DOI: 10.1007/978-1-4419-7747-2_8
    """
    def __init__(self, mate_func, mutate_func, select, create_func):
        super().__init__(mate_func, mutate_func, select)
        self.num_new_blood = 1
        self.create = create_func

    def _init(self, population):
        super()._init(population)
        self.offspring_size -= self.num_new_blood
        self.n_obj = len(population[0].fitness.values)

    def evolve(self, population):
        if not _all_valid(population):
            raise RuntimeError('Cannot evolve on invalid fitness values in population.')
        if not self._initialized:
            self._init(population)
        self._aging(population)
        parents = self.select(population, self.parents_size)[:]
        offspring = self._breed(parents)
        self._remove_age_from_fitness(parents + offspring)
        return parents + offspring + self.create(self.num_new_blood)

    def _remove_age_from_fitness(self, pop):
        for ind in pop:
            ind.fitness.values = ind.fitness.values[:self.n_obj]

    @staticmethod
    def _aging(population):
        for ind in population:
            ind.age = getattr(ind, "age", 0) + 1
            ind.fitness.values = *ind.fitness.values, ind.age


def make_unique_version(obj):
    """Takes an algorithm class and creates a sublcass with a modified evolve method.
    The modified version will ensures uniqueness of individuals.
    """
    uname = "U{}".format(obj.__name__)
    uobj = type(uname, (obj,), {})

    def evolve(self, population):
        return list(set(super().evolve(population)))

    uobj.evlolve = evolve
    return uobj


basic = (NSGA2, SPEA2, DeapEaSimple, AgeFitness)
uniques = []

current_module = sys.modules[__name__]
for alg in basic:
    ualg = make_unique_version(alg)
    setattr(current_module, ualg.__name__, ualg)  # we need to create it in this submodules scope as well
    uniques.append(ualg)

all_algorithms = list(itertools.chain.from_iterable([basic, uniques]))
__all__ = [obj.__name__ for obj in all_algorithms]



from .algorithms import *
from .algorithms import all_algorithms
from .breeding import all_crossover, all_mutations
from .constraints import Constraint, PreTest, PreTestService, NonFiniteExpression, constrain, reject_constrain_violation
from .individual import numpy_phenotype, numpy_primitive_set, sympy_phenotype, sympy_primitive_set
from .individual import Individual, NDIndividual



import os

from setuptools import find_packages, setup

import versioneer

NAME = "pyglyph"
DESCRIPTION = "Symbolic regression tools."
URL = "https://github.com/Ambrosys/glyph"
EMAIL = "markus.quade@ambrosys.de"
AUTHOR = "Markus Abel, Julien Gout, Markus Quade"
KEYWORDS = "complex systems, control, machine learning, genetic programming"
LICENCE = "LGPL"
PYTHON = ">=3.6"

here = os.path.abspath(os.path.dirname(__file__))

with open(os.path.join(here, "requirements.txt"), "r") as f:
    REQUIRED = f.readlines()

REQUIRED_GUI = ["gooey>=1.0.0"]

with open(os.path.join(here, "README.md"), "r") as f:
    LONG_DESCRIPTION = f.read()

if __name__ == "__main__":
    setup(
        name=NAME,
        version=versioneer.get_version(),
        author=AUTHOR,
        author_email=EMAIL,
        description=DESCRIPTION,
        long_description=LONG_DESCRIPTION,
        long_description_content_type="text/markdown",
        license=LICENCE,
        keywords=KEYWORDS,
        url=URL,
        packages=find_packages(exclude=["tests", "doc", "examples"]),
        install_requires=REQUIRED,
        python_requires=PYTHON,
        classifiers=[
            "Development Status :: 3 - Alpha",
            "Topic :: Scientific/Engineering :: Artificial Intelligence",
            "License :: OSI Approved :: GNU Library or Lesser General Public License (LGPL)",
            "Intended Audience :: Science/Research",
            "Programming Language :: Python :: 3 :: Only",
            "Programming Language :: Python :: 3.6",
            "Programming Language :: Python :: 3.7",
        ],
        extras_require={"gui": REQUIRED_GUI},
        entry_points={
            "console_scripts": [
                "glyph-remote = glyph.cli.glyph_remote:main"
            ]
        },
        cmdclass=versioneer.get_cmdclass(),
    )



import logging

import matplotlib
import matplotlib.pyplot as plt
import numpy as np

logger = logging.getLogger(__name__)


def get_limits(x, factor=1.1):
    """Calculates the plot range given an array x."""
    avg = np.nanmean(x)
    range_ = np.nanmax(x) - np.nanmin(x)
    if range_ == 0:
        range_ = 0.5
    return avg - range_/2 * factor, avg + range_/2 * factor


class ProgressObserver(object):  # pragma: no cover
    def __init__(self):
        """Animates the progress of the evolutionary optimization.

        Note:
            Uses matplotlib's interactive mode.
        """
        logger.debug("The ProgressObserver needs an interactive matplotlib backend.")
        logger.debug(f"Using {matplotlib.rcParams['backend']} as backend in matplotlib.")
        logger.debug("Try export MPLBACKEND='TkAgg'")
        plt.ion()
        self.fig = None
        self.axis = None
        self.lines = []
        self.t_key = "gen"

    @staticmethod
    def _update_plt(ax, line, *data):
        x, y = data
        x = np.array(x)
        y = np.array(y)
        ax.set_xlim(*get_limits(x))
        ax.set_ylim(*get_limits(y))
        line.set_data(data)

    def _blank_canvas(self, chapters):
        self.fig, self.axes = plt.subplots(nrows=len(chapters) + 1)
        for c, ax in zip(chapters, self.axes):
            ax.set_xlabel(self.t_key)
            ax.set_ylabel(c)
            ax.set_title("Best " + c)
            line, = ax.plot([], [])
            self.lines.append(line)
        line, = self.axes[-1].plot([], [],  "o-")
        self.axes[-1].set_xlabel(chapters[0])
        self.axes[-1].set_ylabel(chapters[1])
        self.axes[-1].set_title("Pareto Front")
        self.lines.append(line)

        plt.tight_layout()

    def __call__(self, app):
        """
        Note:
            To be used as a callback in :class:`glyph.application.Application`.

        Args:
            app (glyph.application.Application)
        """
        chapters = sorted(app.logbook.chapters.keys())

        if self.fig is None:
            self._blank_canvas(chapters)

        t = app.logbook.select(self.t_key)
        for c, ax, line in zip(chapters, self.axes, self.lines):
            self._update_plt(ax, line, t, app.logbook.chapters[c].select("min"))

        x, y = zip(*sorted([i.fitness.values for i in app.gp_runner.pareto_front]))
        self._update_plt(self.axes[-1], self.lines[-1], x, y)

        self.fig.canvas.draw()



import argparse
import os
import logging

import numpy as np

import glyph.application
import glyph.assessment
from glyph.utils.argparse import (
    positive_int,
    non_negative_int,
    np_infinity_int,
    readable_file,
    readable_yaml_file,
)

logger = logging.getLogger(__name__)

try:
    import gooey
    from gooey import Gooey, GooeyParser

    @Gooey(
        auto_start=False,
        advanced=True,
        encoding="utf-8",
        language="english",
        show_config=True,
        default_size=(1200, 1000),
        dump_build_config=False,
        load_build_config=None,
        monospace_display=False,
        disable_stop_button=False,
        show_stop_warning=True,
        force_stop_is_error=True,
        show_success_modal=True,
        run_validators=True,
        poll_external_updates=False,
        return_to_config=False,
        disable_progress_bar_animation=False,
        navigation="SIDEBAR",
        tabbed_groups=True,
        navigation_title="Actions",
        show_sidebar=False,
        progress_regex=r"^.*INFO\D+\d+\D+(?P<gen>[0-9]+)\D+\d+[.]{1}\d+\D+\d+[.]{1}\d+.*$",
        progress_expr="(gen + 1) % 10 / 10 * 100",
    )
    def get_gooey(prog="glyph-remote"):
        probably_fork = "site-packages" not in gooey.__file__
        logger.debug("Gooey located at {}.".format(gooey.__file__))
        if not probably_fork:
            logger.warning("GUI input validators may have no effect")
        parser = GooeyParser(prog=prog)
        return parser

    GUI_AVAILABLE = True
except ImportError as e:
    logger.error(e)
    GUI_AVAILABLE = False
    GUI_UNAVAILABLE_MSG = """Could not start gui extention.
You need to install the gui extras.
Use the command 'pip install glyph[gui]' to do so."""


class MyGooeyMixin:
    def add_argument(self, *args, **kwargs):
        for key in ["widget"]:
            if key in kwargs:
                del kwargs[key]
        super().add_argument(*args, **kwargs)

    def add_mutually_exclusive_group(self, *args, **kwargs):
        group = MutuallyExclusiveGroup(self, *args, **kwargs)
        self._mutually_exclusive_groups.append(group)
        return group

    def add_argument_group(self, *args, **kwargs):
        group = ArgumentGroup(self, *args, **kwargs)
        self._action_groups.append(group)
        return group


class Parser(MyGooeyMixin, argparse.ArgumentParser):
    pass


class ArgumentGroup(MyGooeyMixin, argparse._ArgumentGroup):
    pass


class MutuallyExclusiveGroup(MyGooeyMixin, argparse._MutuallyExclusiveGroup):
    pass


def get_parser(parser=None):
    if parser is None:
        parser = Parser()
    if isinstance(parser, Parser):
        parser.add_argument("--gui", action="store_true", default=False)

    gui_active = GUI_AVAILABLE and isinstance(parser, GooeyParser)

    parser.add_argument(
        "--port",
        type=positive_int,
        default=5555,
        help="Port for the zeromq communication (default: 5555)",
    )
    parser.add_argument("--ip", type=str, default="localhost", help="IP of the client (default: localhost)")
    parser.add_argument(
        "--send_meta_data", action="store_true", default=False, help="Send metadata after each generation"
    )
    parser.add_argument(
        "-v",
        "--verbose",
        type=str.upper,
        dest="verbosity",
        choices=["CRITICAL", "ERROR", "WARNING", "INFO", "DEBUG", "NOTSET"],
        default="INFO",
        help="Set logging level",
    )
    parser.add_argument(
        "--logging",
        "-l",
        dest="logging_config",
        type=str,
        default="logging.yaml",
        help="set config file for logging; overrides --verbose (default: logging.yaml)",
        widget="FileChooser",
    )

    config = parser.add_argument_group("config")
    group = config.add_mutually_exclusive_group(
        required=gui_active
    )
    group.add_argument(
        "--remote",
        action="store_true",
        dest="remote",
        default=False,
        help="Request GP configs from experiment handler.",
    )
    group.add_argument(
        "--cfile",
        dest="cfile",
        type=readable_yaml_file,
        help="Read GP configs from file",
        widget="FileChooser",
    )

    glyph.application.Application.add_options(parser)
    cp_group = parser.add_mutually_exclusive_group(
        required=gui_active
    )
    cp_group.add_argument("--ndim", type=positive_int, default=1)
    cp_group.add_argument(
        "--resume",
        dest="resume_file",
        metavar="FILE",
        type=readable_file,
        help="continue previous run from a checkpoint file",
        widget="FileChooser",
    )
    cp_group.add_argument(
        "-o",
        dest="checkpoint_file",
        metavar="FILE",
        type=str,
        default=os.path.join(".", "checkpoint.pickle"),
        help="checkpoint to FILE (default: ./checkpoint.pickle)",
        widget="FileChooser",
    )

    glyph.application.AlgorithmFactory.add_options(parser.add_argument_group("algorithm"))
    group_breeding = parser.add_argument_group("breeding")
    glyph.application.MateFactory.add_options(group_breeding)
    glyph.application.MutateFactory.add_options(group_breeding)
    glyph.application.SelectFactory.add_options(group_breeding)
    glyph.application.CreateFactory.add_options(group_breeding)

    ass_group = parser.add_argument_group("assessment")
    ass_group.add_argument(
        "--simplify",
        action="store_true",
        default=False,
        help="Simplify expression before sending them. (default: False)",
    )
    ass_group.add_argument(
        "--complexity_measure",
        choices=["None"] + list(glyph.assessment.complexity_measures.keys()),
        default=None,
        help="Consider the complexity of solutions for MOO (default: None)",
    )
    ass_group.add_argument(
        "--no_caching",
        dest="caching",
        action="store_false",
        default=True,
        help="Cache evaluation (default: False)",
    )
    ass_group.add_argument(
        "--persistent_caching",
        default=None,
        help="Key for persistent data base cache for caching between experiments (default: None)",
    )
    ass_group.add_argument(
        "--max_fev_const_opt",
        type=non_negative_int,
        default=100,
        help="Maximum number of function evaluations for constant optimization (default: 100)",
    )
    ass_group.add_argument(
        "--directions",
        type=positive_int,
        default=5,
        help="Directions for the stochastic hill-climber (default: 5 only used in conjunction with --const_opt_method hill_climb)",
    )
    ass_group.add_argument(
        "--precision",
        type=non_negative_int,
        default=3,
        help="Precision of constants (default: 3)",
    )
    ass_group.add_argument(
        "--const_opt_method",
        choices=["hill_climb", "Nelder-Mead"],
        default="Nelder-Mead",
        help="Algorithm to optimize constants given a structure (default: Nelder-Mead)",
    )
    ass_group.add_argument(
        "--structural_constants",
        action="store_true",
        default=False,
        help="Make use of structural constants. (default: False)",
    )
    ass_group.add_argument(
        "--sc_min", type=float, default=-1, help="Minimum value of sc for scaling. (default: -1)"
    )
    ass_group.add_argument(
        "--sc_max", type=float, default=1, help="Maximum value of sc for scaling. (default: 1)"
    )
    ass_group.add_argument(
        "--smart", action="store_true", default=False, help="Use smart constant optimization. (default: False)"
    )
    ass_group.add_argument(
        "--smart_step_size",
        type=non_negative_int,
        default=10,
        help="Number of fev in iterative function optimization. (default: 10)",
    )
    ass_group.add_argument(
        "--smart_min_stat",
        type=non_negative_int,
        default=10,
        help="Number of samples required prior to stopping (default: 10)",
    )
    ass_group.add_argument(
        "--smart_threshold",
        type=non_negative_int,
        default=25,
        help="Quantile of improvement rate. Abort constant optimization if below (default: 25)",
    )
    ass_group.add_argument(
        "--chunk_size",
        type=positive_int,
        default=30,
        help="Number of individuals send per single request. (default: 30)",
    )
    ass_group.add_argument(
        "--multi_objective",
        action="store_true",
        default=False,
        help="Returned fitness is multi-objective (default: False)",
    )
    ass_group.add_argument(
        "--send_symbolic",
        action="store_true",
        default=False,
        help="Send the expression with symbolic constants (default: False)",
    )
    ass_group.add_argument(
        "--re_evaluate",
        action="store_true",
        default=False,
        help="Re-evaluate old individuals (default: False)",
    )

    break_condition = parser.add_argument_group("break condition")
    break_condition.add_argument(
        "--ttl",
        type=int,
        default=-1,
        help="Time to life (in seconds) until soft shutdown. -1 = no ttl (default: -1)",
    )
    break_condition.add_argument(
        "--target", type=float, default=0, help="Target error used in stopping criteria (default: 0)"
    )
    break_condition.add_argument(
        "--max_iter_total",
        type=np_infinity_int,
        default=np.infty,
        help="Maximum number of function evaluations (default: 'inf' [stands for np.infty])",
    )

    constraints = parser.add_argument_group("constraints")
    glyph.application.ConstraintsFactory.add_options(constraints)

    observer = parser.add_argument_group("observer")
    observer.add_argument(
        "--animate",
        action="store_true",
        default=False,
        help="Animate the progress of evolutionary optimization. (default: False)",
    )

    return parser
